/**
 * CMI : Cluster Method Invocation
 * Copyright (C) 2008-2009 Bull S.A.S.
 * Contact: cmi@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * --------------------------------------------------------------------------
 * $Id: JMSClusterViewManager.java 2438 2010-02-24 16:20:30Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.ow2.cmi.controller.server.impl.jms;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.ObjectStreamClass;
import java.io.Serializable;
import java.rmi.RemoteException;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentHashMap;

import javax.jms.BytesMessage;
import javax.jms.Connection;
import javax.jms.ConnectionFactory;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageConsumer;
import javax.jms.MessageListener;
import javax.jms.MessageProducer;
import javax.jms.Session;
import javax.jms.TextMessage;
import javax.jms.Topic;
import javax.management.MBeanServerNotification;
import javax.management.Notification;
import javax.management.NotificationListener;
import javax.management.ObjectName;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import net.jcip.annotations.ThreadSafe;

import org.ow2.cmi.admin.MBeanUtils;
import org.ow2.cmi.controller.provider.ClusterViewProvider;
import org.ow2.cmi.controller.provider.ServerClusterViewProvider;
import org.ow2.cmi.controller.server.AbsServerClusterViewManager;
import org.ow2.cmi.controller.server.ClusterView;
import org.ow2.cmi.controller.server.DistributedObjectInfo;
import org.ow2.cmi.controller.server.Domain;
import org.ow2.cmi.controller.server.ServerClusterViewManagerException;
import org.ow2.cmi.controller.server.ServerView;
import org.ow2.cmi.controller.server.filter.IFilter;
import org.ow2.cmi.lb.ILoadBalancerTweaker;
import org.ow2.cmi.lb.data.PolicyData;
import org.ow2.cmi.reference.CMIReference;
import org.ow2.cmi.reference.ObjectNotFoundException;
import org.ow2.cmi.reference.ServerId;
import org.ow2.cmi.reference.ServerNotFoundException;
import org.ow2.cmi.reference.ServerRef;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * A JMS-based server cluster view manager.
 * TODO Simultaneously support of many protocols.
 * TODO Register MBean to specify a cluster name.
 * @author Loris Bouzonnet
 */
@ThreadSafe
public final class JMSClusterViewManager extends AbsServerClusterViewManager
implements MessageListener, NotificationListener {

    /**
     * Logger.
     */
    private static Log logger =
        LogFactory.getLog(JMSClusterViewManager.class);

    private static final String PROVIDER_URL_PROPERTY = "providerUrl";

    private static final String ACTION_PROPERTY = "action";

    private static final String TYPE_PROPERTY = "type";

    private static final String ADD_ACTION = "add";

    private static final String REM_ACTION = "rem";

    private static final String SET_ACTION = "set";

    private static final String CMIIREF_TYPE = "cmiref";

    private static final String OBJINFO_TYPE = "objinfo";

    private static final String DELAY_TYPE = "delay";

    private static final String FILTER_TYPE = "filter";

    private static final String SERVER_TYPE = "server";

    private static final String VALUE_PROPERTY = "value";

    private static final String BLACKLIST_TYPE = "blacklist";

    private static final String LOAD_TYPE = "load";

    /**
     * Topic.
     */
    private Topic topic;

    /**
     * Subscriber for the topic.
     */
    private MessageConsumer subscriber;

    /**
     * Publisher for the topic.
     */
    private MessageProducer publisher;

    /**
     * Connection for the topic.
     */
    private Connection topicConnection;

    /**
     * Subscriber session for the topic.
     */
    private Session sessionSubscriber;

    /**
     * Publisher session for the topic.
     */
    private Session sessionPublisher;

    /**
     * local copy of the cluser view.
     */
    private ClusterView clusterView = new ClusterView();

    /**
     * Delegate to register a listener for being notified of the availability of JMS.
     */
    private ObjectName delegate;

    /**
     * The ObjectName of a MBean indicating the availability of JMS.
     */
    private ObjectName objectNameToListen;

    /**
     * FIXME To prevent a simultaneously use of many protocols.
     */
    private ServerId localRegistryId;

    /**
     * Periodically clean the cluster view.
     */
    private ClusterViewCleaner clusterViewCleaner;

    /**
     * Heart beat to notify the other servers that I stay alive.
     */
    private Heartbeat heartbeat;

    /**
     * Embedded JORAM server.
     */
    private JORAMServer joramServer;


    @Override
    public void doStart() throws JMSClusterViewManagerException {

        // Check the configuration
        if (isDistributed() &&
                (getExtTopicJNDIName() == null || getLocalTopicJNDIName() == null)) {
            logger.error("Invalid configuration: in a distributed config, the JNDI names have to be specified for both the local and external topic");
            throw new JMSClusterViewManagerException(
                    "Invalid configuration: in a distributed config, the JNDI names have to be specified for both the local and external topic");
        }

        if (!isDistributed()) {
            if (getExtTopicJNDIName() == null && getLocalTopicJNDIName() == null) {
                logger.error("Invalid configuration: no topic JNDI name specified");
                throw new JMSClusterViewManagerException("Invalid configuration: no topic JNDI name specified");
            }
            if (getExtTopicJNDIName() != null && getLocalTopicJNDIName() != null) {
                logger.error("Invalid configuration: in a non distributed config, the JNDI names have to be specified for either the local topic or the external topic");
                throw new JMSClusterViewManagerException(
                        "Invalid configuration: in a non distributed config, the JNDI names have to be specified for either the local topic or the external topic");
            }
        }

        if (!isSubscriber() && !isPublisher()) {
            logger.error("Invalid configuration: neither a subscriber and nor a publisher");
            throw new JMSClusterViewManagerException("Invalid configuration: neither a subscriber and nor a publisher");
        }

        // Choose of a JMS server
        if(((Config) getConfig()).isJmsServerEmbedded()) {

            // Launch an embedded JMS server according to the configuration.
            joramServer = new JORAMServer((Config) getConfig());
            try {
                joramServer.start();
            } catch (JORAMServerException e) {
                logger.error("Unable to start an embedded instance of JORAM server", e);
                throw new JMSClusterViewManagerException(
                        "Unable to start an embedded instance of JORAM server", e);
            }
            setRegistrationOpened(true);

        } else {

            // Reuse an existing JMS server
            if(getObjectNameToListen() != null) {
                // Its availability is given by a MBean
                try {
                    objectNameToListen = new ObjectName(getObjectNameToListen());
                } catch (Exception e) {
                    logger.error("Unable to check if JMS is available", e);
                    throw new JMSClusterViewManagerException(
                            "Unable to check if JMS is available", e);
                }
            }
            if(getObjectNameToListen() == null
                    || MBeanUtils.getMBeanServer().isRegistered(objectNameToListen)) {

                // The JMS server is available or none ObjectName are provided to check it.
               setRegistrationOpened(true);

            } else {

                // The JMS server is not yet available, waiting for it.
                // Register a listener to be notified of the availability of JMS.
                try {
                    delegate = new ObjectName("JMImplementation:type=MBeanServerDelegate");
                    MBeanUtils.getMBeanServer().addNotificationListener(
                            delegate, this, null, null);
                } catch (Exception e) {
                    logger.error("Unable to add the listener for JMS", e);
                    throw new JMSClusterViewManagerException(
                            "Unable to add the listener for JMS", e);
                }
                logger.info("CMI server waiting for a JMS implementation...");
            }
        }
    }


    @Override
    public void doStop() throws JMSClusterViewManagerException {
        removeLocalServerView();
        setRegistrationOpened(false);
        setImplementationAvailable(false);
        closeConnection();
        clusterView.clear();

        if(delegate != null) {
            if(MBeanUtils.getMBeanServer().isRegistered(delegate)) {
                try {
                    MBeanUtils.getMBeanServer().removeNotificationListener(
                            delegate, this);
                } catch (Exception e) {
                    logger.warn("Unable to remove the listener for JMS", e);
                }
            }
        }

        if (joramServer != null) {
            joramServer.stop();
        }
    }


    /**
     * @return
     */
    private boolean isDistributed() {
        return isClustered() || isHierarchical();
    }


    public void handleNotification(final Notification notification, final Object handback) {
        if (notification.getType().equals(
                MBeanServerNotification.REGISTRATION_NOTIFICATION)) {
            if(notification instanceof MBeanServerNotification
                    && objectNameToListen.equals(((MBeanServerNotification) notification).getMBeanName())) {

                setRegistrationOpened(true);
                if(getState().equals(State.STARTED)) {
                    try {
                        new InitialContext();
                    } catch (NamingException e) {
                        logger.warn("Unable to create a cmi context", e);
                    }
                }

            }
        } else if(notification.getType().equals(
                MBeanServerNotification.UNREGISTRATION_NOTIFICATION)) {
            if(notification instanceof MBeanServerNotification
                    && objectNameToListen.equals(((MBeanServerNotification) notification).getMBeanName())) {
                logger.warn("JMS no more available: auto-stopping the CMI server.");
                sessionPublisher = null;
                sessionSubscriber = null;
                publisher = null;
                subscriber = null;
                topicConnection = null;
                topic = null;
                stop();
            }
        }
    }

    @Override
    protected void initServerConfig(final String contextFactory,
            final ServerId serverId, final Context cmiContext) throws ServerClusterViewManagerException {

        // FIXME To prevent a simultaneously use of many protocols.
        if(localRegistryId == null) {

            localRegistryId = serverId;

            Short id = ((Config) getConfig()).getServerId();

            if (id == null) {
                logger.error("Invalid configuration: server id missing!");
                throw new JMSClusterViewManagerException(
                        "Invalid configuration: server id missing!");
            }

            // Create the server view associated to this serverId
            ServerView serverView;
            try {
                serverView = new ServerView(localRegistryId, id.shortValue(),
                        JORAMHelper.getServerName());
            } catch (JORAMServerException e) {
                logger.error("Unable to retrieve the local server name", e);
                throw new JMSClusterViewManagerException("Unable to retrieve the local server name", e);
            }
            serverView.addDomain(new Domain(getDomainName(), getDomainPort()));

            // Retrieve an external topic according to the configuration
            Topic extTopic = null;
            Context extContext = null;

            if (getExtProviderURL() != null) {

                // The topic will be retrieved from a specific server
                Hashtable<String, String> env = null;
                env = new Hashtable<String, String>();
                env.put(Context.INITIAL_CONTEXT_FACTORY, contextFactory);
                env.put(Context.PROVIDER_URL, getExtProviderURL());
                try {
                    extContext = new InitialContext(env);
                } catch (NamingException e) {
                    logger.error("Unable to create a context", e);
                    throw new JMSClusterViewManagerException("Unable to create a context", e);
                }
            } else if (isDistributed()) {
                extContext = cmiContext;
            }

            // Try to retrieve an existing cluster view
            try {
                clusterView = retrieveClusterView(cmiContext);
            } catch (Exception e) {
                logger.warn("Cannot retrieve the cluster view: no provider found for the jndi name {0}",
                        getJmsProviderName(), e);
            }

            if (extContext != null) {
                try {
                    try {
                        extTopic = (Topic) extContext.lookup(getExtTopicJNDIName());
                    } catch (NamingException e) {
                        logger.error("Unable to retrieve the external topic with name " + getExtTopicJNDIName(), e);
                        throw new JMSClusterViewManagerException(
                                "Unable to retrieve the external topic with name " + getExtTopicJNDIName(), e);
                    }

                    if (extContext instanceof ILoadBalancerTweaker) {

                        ServerRef lastServerRef = ((ILoadBalancerTweaker) cmiContext).getLastServerRef();
                        ((ILoadBalancerTweaker) cmiContext).setNextServerRef(lastServerRef);
                    }

                    topic = extTopic;

                    ConnectionFactory cf;
                    try {
                        cf = (ConnectionFactory) extContext.lookup(getConnectionFactoryName());
                    } catch (NamingException e) {
                        logger.error("Unable to retrieve the connection factory with name " + getConnectionFactoryName(), e);
                        throw new JMSClusterViewManagerException(
                                "Unable to retrieve the connection factory with name " + getConnectionFactoryName(), e);
                    }

                    initConnection(cf, isDistributed());

                    if (extContext instanceof ILoadBalancerTweaker) {
                        ((ILoadBalancerTweaker) cmiContext).setNextServerRef(null);
                    }

                    // Notify the other servers in order that they will be ready.
                    HashMap<String, Serializable> properties = new HashMap<String, Serializable>();
                    properties.put(ACTION_PROPERTY, ADD_ACTION);
                    properties.put(TYPE_PROPERTY, SERVER_TYPE);
                    sendMessage(serverView, properties);
                } finally {
                    if (getExtProviderURL() != null && extContext != null) {
                        try {
                            extContext.close();
                        } catch (NamingException e) {
                            logger.debug("Unable to close the context", e);
                        }
                    }
                }
            }

            if(isDistributed()) {

                closeConnection();

                // The other are waiting me, I'm coming into the place!
                try {
                    JORAMHelper.setDistributedConfig((Config) getConfig(), clusterView);
                } catch (JORAMServerException e) {
                    logger.error("Unable to set a distributed config", e);
                    throw new JMSClusterViewManagerException("Unable to set a distributed config", e);
                }
            }

            if (getLocalTopicJNDIName() != null) {
                String localProviderURL = serverId.getProviderURL();
                Context localContext = null;

                Hashtable<String, String> env = new Hashtable<String, String>();
                env.put(Context.INITIAL_CONTEXT_FACTORY, contextFactory);
                env.put(Context.PROVIDER_URL, localProviderURL);
                try {
                    localContext = new InitialContext(env);
                } catch (NamingException e) {
                    logger.error("Unable to create a context", e);
                    throw new JMSClusterViewManagerException("Unable to create a context", e);
                }
                try {
                    try {
                        topic = (Topic) localContext.lookup(getLocalTopicJNDIName());
                        logger.debug("Topic found with name {0} in the local registry", getLocalTopicJNDIName());
                    } catch (NamingException e) {
                        logger.debug("No topic found with name {0} in the local registry.",
                                getLocalTopicJNDIName(), e);
                        topic = null;
                    }

                    if (topic == null) {
                        // The topic does not yet exist: creating it
                        try {
                            topic = JORAMHelper.createTopic(
                                    getLocalTopicName(), extTopic, isClustered(), isHierarchical());
                        } catch (Exception e) {
                            logger.error("Cannot create the topic {0}", getLocalTopicName(), e);
                            throw new JMSClusterViewManagerException(
                                    "Cannot create the topic " + getLocalTopicName(), e);
                        }
                        try {
                            localContext.rebind(getLocalTopicJNDIName(), topic);
                            logger.debug("My topic with name {0} has been bound", getLocalTopicJNDIName());
                        } catch (NamingException e) {
                            logger.warn("Unable to bind my topic with name " + getLocalTopicJNDIName());
                        }
                    } else if (isDistributed()) {
                        try {
                            JORAMHelper.checkMyTopic(topic, extTopic, isClustered(), isHierarchical());
                        } catch (JORAMServerException e) {
                            logger.warn("The topic checking has failed", e);
                        }
                    }
                    ConnectionFactory cf;
                    if (joramServer != null) {
                        try {
                            localContext.rebind(getConnectionFactoryName(), joramServer.getTcpConnectionFactory());
                        } catch (NamingException e) {
                            logger.error("Unable to bind the TCP connection factory", e);
                            throw new JMSClusterViewManagerException("Unable to bind the TCP connection factory", e);
                        }
                        cf = joramServer.getLocalConnectionFactory();
                    } else {
                        try {
                            cf = (ConnectionFactory) localContext.lookup(getConnectionFactoryName());
                        } catch (NamingException e) {
                            logger.error("Unable to retrieve the connection factory with name " + getConnectionFactoryName(), e);
                            throw new JMSClusterViewManagerException(
                                    "Unable to retrieve the connection factory with name " + getConnectionFactoryName(), e);
                        }
                    }
                    initConnection(cf, false);
                } finally {
                    try {
                        localContext.close();
                    } catch (NamingException e) {
                        logger.warn("Unable to close the context", e);
                    }
                }
            }

            // Add my-self into the cluster view
            clusterView.addServerView(localRegistryId.getProviderURL(), serverView);

            super.initServerConfig(contextFactory, serverId, cmiContext);

            if(isClusterViewShared()) {
                bindClusterViewProvider(contextFactory, serverId);
            }

            setImplementationAvailable(true);

            logger.info("CMI server available.");

            initHeartBeat();

            if (isSubscriber()) {
                initClusterViewCleaner();
            }
        }
    }


    private void initClusterViewCleaner() {
        clusterViewCleaner = new ClusterViewCleaner();
        getCmiThreadFactory().newThread(clusterViewCleaner, "Cluster View Cleaner").start();
    }


    private void initHeartBeat() {
        heartbeat = new Heartbeat();
        getCmiThreadFactory().newThread(heartbeat, "Heart Beat").start();
    }


    private void bindClusterViewProvider(
            final String contextFactory, final ServerId serverId) throws JMSClusterViewManagerException {

        String myProviderURL = serverId.getProviderURL();
        Context myContext = null;

        Hashtable<String, String> env = new Hashtable<String, String>();
        env.put(Context.INITIAL_CONTEXT_FACTORY, contextFactory);
        env.put(Context.PROVIDER_URL, myProviderURL);
        try {
            myContext = new InitialContext(env);
        } catch (NamingException e) {
            logger.error("Unable to create a context", e);
            throw new JMSClusterViewManagerException("Unable to create a context", e);
        }
        ClusterViewProvider jmsProvider;
        try {
            jmsProvider = new JmsProvider(this);
        } catch (RemoteException e) {
            logger.error("Unable to create the JMS Provider", e);
            throw new JMSClusterViewManagerException("Unable to create the JMS Provider", e);
        }
        try {
            myContext.rebind(getJmsProviderName(), jmsProvider);
            logger.debug("The JMS provider with name {0} has been bound", getJmsProviderName());
        } catch(NamingException e) {
            logger.error("Cannot share the cluster view provider with name " + getJmsProviderName(), e);
            throw new JMSClusterViewManagerException(
                    "Cannot share the cluster view provider with name " + getJmsProviderName(), e);
        }
    }

    private void initConnection(final ConnectionFactory cf, final boolean publisherOnly) throws JMSClusterViewManagerException {

        try {
            topicConnection = cf.createConnection();
            if (isSubscriber() && !publisherOnly) {
                sessionSubscriber = topicConnection.createSession(false, Session.AUTO_ACKNOWLEDGE);
                subscriber = sessionSubscriber.createConsumer(topic);
                subscriber.setMessageListener(this);
            }
            if (isPublisher() || publisherOnly) {
                sessionPublisher = topicConnection.createSession(false, Session.AUTO_ACKNOWLEDGE);
                publisher = sessionPublisher.createProducer(topic);
            }

            // Start the connection
            topicConnection.start();
        } catch (JMSException e) {
            logger.error("Unable to initialize a connection", e);
            throw new JMSClusterViewManagerException("Unable to initialize a connection", e);
        }
    }

    private void closeConnection() throws JMSClusterViewManagerException {

        try {
            if (topicConnection != null) {
                topicConnection.close();
                topicConnection = null;
            }
            if (publisher != null) {
                publisher.close();
                publisher = null;
            }
            if (subscriber != null) {
                subscriber.close();
                subscriber = null;
            }
            if (sessionSubscriber != null) {
                sessionSubscriber.close();
                sessionSubscriber = null;
            }
            if (sessionPublisher != null) {
                sessionPublisher.close();
                sessionPublisher = null;
            }

        } catch (JMSException e) {
            logger.error("Unable to close the connection", e);
            throw new JMSClusterViewManagerException("Unable to close the connection", e);
        }
    }

    private ClusterView retrieveClusterView(final Context cmiContext) throws RemoteException, NamingException {
        ClusterView cView = ((ServerClusterViewProvider)
                cmiContext.lookup(getJmsProviderName())).getClusterView();

        logger.debug("Cluster view retrieved: {0}", clusterView);
        return cView;
    }

    private void removeLocalServerView() {
        ServerId serverId = getLocalServer();
        try {
            removeServerView(serverId.getProviderURL());
        } catch (ServerNotFoundException e) {
            logger.warn("Unable to remove my-self from the cluster view!", e);
        }
        HashMap<String, Serializable> properties = new HashMap<String, Serializable>();
        properties.put(ACTION_PROPERTY, REM_ACTION);
        properties.put(TYPE_PROPERTY, SERVER_TYPE);
        sendMessage(null, properties);
    }

    private void addServerView(final String providerUrl, final ServerView serverView) throws JMSClusterViewManagerException {
        if (clusterView.addServerView(providerUrl, serverView)) {
            if (getDomainName() != null
                    && !localRegistryId.getProviderURL().equals(providerUrl)) {
                try {
                    JORAMHelper.addDomain((Config) getConfig());
                    JORAMHelper.addServer(serverView, getDomainName());
                } catch (JORAMServerException e) {
                    logger.error("Unable to update the distributed config!", e);
                    throw new JMSClusterViewManagerException("Unable to update the distributed config!", e);
                }
            }
        }
    }

    /**
     * @param providerUrl
     * @throws ServerNotFoundException
     */
    private void removeServerView(final String providerUrl) throws ServerNotFoundException {
        clusterView.removeClusteredObjects(providerUrl);
        ServerView serverView = clusterView.removeServerView(providerUrl);
        if(serverView == null) {
            logger.error("Unknown server: " + providerUrl);
            throw new ServerNotFoundException("Unknown server", providerUrl);
        }
        if (getDomainName() != null
                && !localRegistryId.getProviderURL().equals(providerUrl)) {
            Domain domain = serverView.getDomain(getDomainName());
            if (domain != null) {
                try {
                    JORAMHelper.removeServer(topic, serverView.getId());
                } catch (JORAMServerException e) {
                    logger.warn("Unable to update the distributed config!", e);
                }
            }
        }
    }


    @Override
    protected boolean containObject(final String objectName) {
        return clusterView.containsObject(objectName);
    }

    @Override
    protected DistributedObjectInfo getDistributedObjectInfo(final String objectName)
            throws ObjectNotFoundException {
        return clusterView.getObjectInfo(objectName);
    }

    @Override
    public void broadCastArchive(final Object archiveId, final byte[] bytesOfFile) {
        throw new UnsupportedOperationException("TODO");
    }

    @Override
    protected void initStats() {
        // TODO
    }

    @Override
    public boolean isAlreadyDistributed(final Object archiveId) {
        throw new UnsupportedOperationException("TODO");
    }

    @Override
    public void removeDistributedArchive(final Object archiveId) {
        throw new UnsupportedOperationException("TODO");
    }

    private void sendMessage(final Object data, final Map<String, Serializable> properties) {
        if (isPublisher()) {
            if(!(data instanceof Serializable)) {
                throw new IllegalArgumentException("The provided object is not serializable.");
            }
            try {
                Message msg;
                if (data ==  null) {
                    msg = sessionPublisher.createMessage();
                } else {
                    msg = sessionPublisher.createBytesMessage();
                    ((BytesMessage) msg).writeBytes(serialize(data));
                }
                for(Entry<String, Serializable> entry : properties.entrySet()) {
                    msg.setObjectProperty(entry.getKey(), entry.getValue());
                }
                msg.setStringProperty(PROVIDER_URL_PROPERTY, getLocalServer().getProviderURL());
                publisher.send(msg);
            } catch (Exception e) {
                logger.warn("Unable to send a message to add {0} with properties {1}", data, properties, e);
            }
        }
    }

    /**
     * @return
     */
    private ServerId getLocalServer() {
        return localRegistryId;
    }


    @Override
    protected void addDistributedObjectInfo(final String objectName,
            final DistributedObjectInfo distributedObjectInfo) {
        addLocalDistributedObjectInfo(objectName, distributedObjectInfo);
        HashMap<String, Serializable> properties = new HashMap<String, Serializable>();
        properties.put(ACTION_PROPERTY, ADD_ACTION);
        properties.put(TYPE_PROPERTY, OBJINFO_TYPE);
        sendMessage(distributedObjectInfo, properties);
    }


    /**
     * Sets informations on the clustered object with the given name.
     *
     * @param objectName
     *            a name of object
     * @param distributedObjectInfo
     *            informations on the clustered object
     */
    @Override
    protected void setDistributedObjectInfo(final String objectName,
            final DistributedObjectInfo distributedObjectInfo) {
        setLocalDistributedObjectInfo(objectName, distributedObjectInfo);
        HashMap<String, Serializable> properties = new HashMap<String, Serializable>();
        properties.put(ACTION_PROPERTY, SET_ACTION);
        properties.put(TYPE_PROPERTY, OBJINFO_TYPE);
        sendMessage(distributedObjectInfo, properties);
    }

    public void addCMIReference(final CMIReference cmiReference) {
        addLocalCMIReference(cmiReference);
        HashMap<String, Serializable> properties = new HashMap<String, Serializable>();
        properties.put(ACTION_PROPERTY, ADD_ACTION);
        properties.put(TYPE_PROPERTY, CMIIREF_TYPE);
        sendMessage(cmiReference, properties);
    }


    public List<CMIReference> getCMIReferences(final String objectName)
            throws ObjectNotFoundException {
        return clusterView.getCMIReferences(objectName);
    }

    public Collection<CMIReference> getCMIReferences(final String objectName,
            final String protocolName)
            throws ObjectNotFoundException {
        return clusterView.getCMIReferences(objectName, protocolName);
    }

    public Set<String> getObjectNames() {
        return clusterView.getObjectNames();
    }


    private void setLocalDistributedObjectInfo(final String objectName,
            final DistributedObjectInfo distributedObjectInfo) {

        logger.debug("Updating the infos of {0} with {1}", objectName, distributedObjectInfo);

        DistributedObjectInfo oldDistributedObjectInfo =
            clusterView.addObjectInfo(distributedObjectInfo, true);
        if(!oldDistributedObjectInfo.equals(distributedObjectInfo)) {
            PolicyData policyData = oldDistributedObjectInfo.getPolicyData();
            if(!policyData.equals(distributedObjectInfo.getPolicyData())) {
                try {
                    updatePolicy(objectName);
                } catch (Exception e) {
                    logger.error("Cannot update the policy for the object with name " + objectName, e);
                }
            }
            try {
               updatePool(oldDistributedObjectInfo, distributedObjectInfo);
            } catch (ObjectNotFoundException e) {
                logger.error("Cannot update the pool configuration for the object with name " + objectName, e);
            }
        }
    }

    private void addLocalDistributedObjectInfo(final String objectName,
            final DistributedObjectInfo distributedObjectInfo) {

        logger.debug("Adding the infos of {0} with {1}", objectName, distributedObjectInfo);

        clusterView.addObjectInfo(distributedObjectInfo, false);
        try {
            updatePolicy(objectName);
        } catch (Exception e) {
            logger.warn("Cannot update the policy for the object with name " + objectName, e);
        }
        try {
            updatePool(null, distributedObjectInfo);
        } catch (ObjectNotFoundException e) {
            logger.warn("Cannot update the pool configuration for the object with name " + objectName, e);
        }
    }

    private void addLocalCMIReference(final CMIReference cmiReference) {
        clusterView.addCMIReference(cmiReference);
    }

    @Override
    protected boolean doRemoveCMIReference(final CMIReference cmiReference) {
        try {
            removeLocalCMIReference(cmiReference);
            HashMap<String, Serializable> properties = new HashMap<String, Serializable>();
            properties.put(ACTION_PROPERTY, REM_ACTION);
            properties.put(TYPE_PROPERTY, CMIIREF_TYPE);
            sendMessage(cmiReference, properties);
            if(clusterView.getCMIReferences(cmiReference.getObjectName()).isEmpty()) {
                clusterView.removeCMIReference(cmiReference);
                return true;
            }
        } catch (ObjectNotFoundException e) {
            logger.debug("Cannot remove the CMIReference {0}", cmiReference, e);
        }
        return false;
    }

    private void removeLocalCMIReference(final CMIReference cmiReference) throws ObjectNotFoundException {
        clusterView.removeCMIReference(cmiReference);
    }

    public void setDelayToRefresh(final int delay) {
        setLocalDelay(delay);
        HashMap<String, Serializable> properties = new HashMap<String, Serializable>();
        properties.put(ACTION_PROPERTY, SET_ACTION);
        properties.put(TYPE_PROPERTY, DELAY_TYPE);
        properties.put(VALUE_PROPERTY, delay);
        sendMessage(null, properties);
    }

    /* (non-Javadoc)
     * @see org.ow2.cmi.controller.server.AbsServerClusterViewManager#initDelayToRefresh(int)
     */
    @Override
    protected void initDelayToRefresh(final int delayToRefresh) {
        // Do nothing
    }


    private void setLocalDelay(final int delay) {
        clusterView.setDelay(delay);
    }


    public int getDelayToRefresh() {
        return clusterView.getDelay();
    }

    public List<IFilter> getGlobalFilters() {
        return clusterView.getFilters();
    }

    public void addGlobalFilter(final IFilter filter) throws IllegalArgumentException {
        addFilter(filter);
        HashMap<String, Serializable> properties = new HashMap<String, Serializable>();
        properties.put(ACTION_PROPERTY, ADD_ACTION);
        sendMessage(filter, properties);
    }

    public void removeGlobalFilter(final IFilter filter) {
        if(removeFilter(filter)) {
            HashMap<String, Serializable> properties = new HashMap<String, Serializable>();
            properties.put(ACTION_PROPERTY, REM_ACTION);
            sendMessage(filter, properties);
        }
    }

    private void addFilter(final IFilter filter) {
        clusterView.addFilter(filter);
    }

    private boolean removeFilter(final IFilter filter) {
        return clusterView.removeFilter(filter);
    }


    public boolean isServerBlackListed(final ServerRef serverRef) {
        return clusterView.isServerBlackListed(serverRef.getProviderURL());
    }

    public void addServerToBlackList(final ServerRef serverRef) {
        try {
            if(!clusterView.addServerToBlackList(serverRef.getProviderURL())) {
                HashMap<String, Serializable> properties = new HashMap<String, Serializable>();
                properties.put(ACTION_PROPERTY, SET_ACTION);
                properties.put(TYPE_PROPERTY, BLACKLIST_TYPE);
                properties.put(VALUE_PROPERTY, true);
                sendMessage(null, properties);
            }
        } catch (ServerNotFoundException e) {
            logger.warn("Unable to add to the blacklist since unknown server: {0}", serverRef, e);
        }
    }

    public void removeServerFromBlackList(final ServerRef serverRef) {
        try {
            if(clusterView.removeServerFromBlackList(serverRef.getProviderURL())) {
                HashMap<String, Serializable> properties = new HashMap<String, Serializable>();
                properties.put(ACTION_PROPERTY, SET_ACTION);
                properties.put(TYPE_PROPERTY, BLACKLIST_TYPE);
                properties.put(VALUE_PROPERTY, false);
                sendMessage(null, properties);
            }
        } catch (ServerNotFoundException e) {
            logger.warn("Unable to remove from the blacklist since unknown server: {0}", serverRef, e);
        }
    }

    public Set<String> getClusterNames() {
        return clusterView.getClusterNames();
    }


    public int getNbClientsConnectedToProvider() {
        return -1;
    }

    public Set<String> getObjectNames(final String clusterName) {
        return clusterView.getObjectNames(clusterName);
    }

    public void registerClient(final UUID uuid) {
        // TODO
    }


    public boolean isPoolToEmpty(final String objectName)
            throws ObjectNotFoundException {
        throw new UnsupportedOperationException("TODO");
    }

    public void addPoolToEmpty(final String objectName) {
        throw new UnsupportedOperationException("TODO");
    }

    public void removePoolToEmpty(final String objectName) {
        throw new UnsupportedOperationException("TODO");
    }

    public int getLoadFactor(final ServerRef serverRef)
            throws ServerNotFoundException {
        return clusterView.getLoadFactor(serverRef.getProviderURL());
    }

    public void setLoadFactor(final ServerRef serverRef, final int loadFactor) {
        try {
            if(clusterView.setLoadFactor(serverRef.getProviderURL(), loadFactor) != loadFactor) {
                HashMap<String, Serializable> properties = new HashMap<String, Serializable>();
                properties.put(ACTION_PROPERTY, SET_ACTION);
                properties.put(TYPE_PROPERTY, LOAD_TYPE);
                properties.put(VALUE_PROPERTY, loadFactor);
                sendMessage(null, properties);
            }
        } catch (ServerNotFoundException e) {
            logger.warn("Unable to set the loadfactor since unknown server: {0}", serverRef, e);
        }
    }

    public long getDateOfConfiguration() {
        // TODO
        return -1;
    }

    public void onMessage(final Message msg) {

        // Retrieve properties
        String providerURL;
        try {
            providerURL = msg.getStringProperty(PROVIDER_URL_PROPERTY);
        } catch (JMSException e) {
            logger.warn("Unable to retrieve the provider URL: ignoring the message {0}", msg, e);
            return;
        }

        if(providerURL == null) {
            logger.warn("Unable to retrieve the provider URL: ignoring the message {0}", msg);
            return;
        }

        String action;
        try {
            action = msg.getStringProperty(ACTION_PROPERTY);
        } catch (JMSException e) {
            logger.warn("Unable to retrieve the action property: ignoring the message {0}", msg);
            return;
        }
        String type;
        try {
            type = msg.getStringProperty(TYPE_PROPERTY);
        } catch (JMSException e) {
            logger.warn("Unable to retrieve the type property: ignoring the message {0}", msg);
            return;
        }

        Object value = null;

        // Retrieve content
        if(msg instanceof BytesMessage) {
            BytesMessage bytesMessage = (BytesMessage) msg;
            try {
                value = unserialize(bytesMessage);
            } catch (Exception e) {
                logger.error("Cannot unserialize the bytes message {0}", bytesMessage, e);
                return;
            }
        } else {
            try {
                value = getValue(msg);
            } catch (JMSException e) {
                logger.warn("Unable to retrieve the value: ignoring the message {0}", msg);
                return;
            }
        }

        if(!providerURL.equals(getLocalServer().getProviderURL())) {
            try {
                process(providerURL, action, type, value);
            } catch (ServerClusterViewManagerException e) {
                logger.warn("The processing has failed for the message: {0}", msg);
                return;
            }
        }
    }

    private void process(final String providerURL, final String action, final String type, final Object value)
    throws ServerClusterViewManagerException {

        if (action.equals(ADD_ACTION)) {
            if (type.equals(CMIIREF_TYPE)) {
                if (!(value instanceof CMIReference)) {
                    logger.error("Illegal value {0} for the specified type {1}", value, type);
                    throw new ServerClusterViewManagerException("Illegal value "
                            + value + " for the specified type " + type);
                }
                logger.debug("A new object is available at this location: {0}", value);
                addLocalCMIReference((CMIReference) value);
            } else if(type.equals(OBJINFO_TYPE)) {
                if(!(value instanceof DistributedObjectInfo)) {
                    logger.error("Illegal value {0} for the specified type {1}", value, type);
                    throw new ServerClusterViewManagerException("Illegal value "
                            + value + " for the specified type " + type);
                }
                addLocalDistributedObjectInfo(
                        ((DistributedObjectInfo) value).getObjectName(), (DistributedObjectInfo) value);
            } else if(type.equals(FILTER_TYPE)) {
                if(!(value instanceof IFilter)) {
                    logger.error("Illegal value {0} for the specified type {1}", value, type);
                    throw new ServerClusterViewManagerException("Illegal value "
                            + value + " for the specified type " + type);
                }
                addFilter((IFilter) value);
            } else if (type.equals(SERVER_TYPE)) {
                if (!(value instanceof ServerView)) {
                    logger.error("Illegal value {0} for the specified type {1}", value, type);
                    throw new ServerClusterViewManagerException("Illegal value "
                            + value + " for the specified type " + type);
                }
                addServerView(providerURL, (ServerView) value);
            } else {
                logger.error("Unknown type {1} for the action {0}", action, type);
                throw new ServerClusterViewManagerException("Unknown type "
                        + type + " for the action " + action);
            }
        } else if (action.equals(REM_ACTION)) {
            if (type.equals(CMIIREF_TYPE)) {
                if (!(value instanceof CMIReference)) {
                    logger.error("Illegal value {0} for the specified type {1}", value, type);
                    throw new ServerClusterViewManagerException("Illegal value "
                            + value + " for the specified type " + type);
                }
                logger.debug("A new object is to remove for this location: {0}", value);
                try {
                    removeLocalCMIReference((CMIReference) value);
                } catch (ObjectNotFoundException e) {
                    logger.error("Cannot remove the CMIReference {0}", value, e);
                    throw new ServerClusterViewManagerException("Cannot remove the CMIReference " + value, e);
                }
            } else if(type.equals(FILTER_TYPE)) {
                if(!(value instanceof IFilter)) {
                    logger.error("Illegal value {0} for the specified type {1}", value, type);
                    throw new ServerClusterViewManagerException("Illegal value "
                            + value + " for the specified type " + type);
                }
                removeFilter((IFilter) value);
            } else if(type.equals(SERVER_TYPE)) {
                try {
                    removeServerView(providerURL);
                } catch (ServerNotFoundException e) {
                    logger.error("Unknown server: {0}", value, e);
                    throw new ServerClusterViewManagerException("Unknown server: " + value, e);
                }
            } else {
                logger.error("Unknown type {1} for the action {0}", action, type);
                throw new ServerClusterViewManagerException("Unknown type "
                        + type + " for the action " + action);
            }
        } else if (action.equals(SET_ACTION)) {
            if(value instanceof DistributedObjectInfo) {
                setLocalDistributedObjectInfo(
                        ((DistributedObjectInfo) value).getObjectName(), (DistributedObjectInfo) value);
            } else if (type.equals(BLACKLIST_TYPE)) {
                if(!(value instanceof Boolean)) {
                    logger.error("Illegal value {0} for the specified type {1}", value, type);
                    throw new ServerClusterViewManagerException("Illegal value "
                            + value + " for the specified type " + type);
                }
                if((Boolean) value) {
                    try {
                        clusterView.addServerToBlackList(providerURL);
                    } catch (ServerNotFoundException e) {
                        logger.error("Unknown server: {0}", value, e);
                        throw new ServerClusterViewManagerException("Unknown server: " + value, e);
                    }
                } else {
                    try {
                        clusterView.removeServerFromBlackList(providerURL);
                    } catch (ServerNotFoundException e) {
                        logger.error("Unknown server: {0}", value, e);
                        throw new ServerClusterViewManagerException("Unknown server: " + value, e);
                    }
                }
            } else if (type.equals(LOAD_TYPE)) {
                if(!(value instanceof Integer)) {
                    logger.error("Illegal value {0} for the specified type {1}", value, type);
                    throw new ServerClusterViewManagerException("Illegal value "
                            + value + " for the specified type " + type);
                }
                try {
                    clusterView.setLoadFactor(providerURL, (Integer) value);
                } catch (ServerNotFoundException e) {
                    logger.error("Unknown server: {0}", value, e);
                    throw new ServerClusterViewManagerException("Unknown server: " + value, e);
                }
            } else if (type.equals(DELAY_TYPE)) {
                if(!(value instanceof Integer)) {
                    logger.error("Illegal value {0} for the specified type {1}", value, type);
                    throw new ServerClusterViewManagerException("Illegal value "
                            + value + " for the specified type " + type);
                }
                setLocalDelay((Integer) value);
            } else {
                logger.error("Unknown type {1} for the action {0}", action, type);
                throw new ServerClusterViewManagerException("Unknown type "
                        + type + " for the action " + action);
            }
        } else {
            logger.error("Unknown action: {0}", action);
            throw new ServerClusterViewManagerException("Unknown action: " + action);
        }
    }

    private Object getValue(final Message msg) throws JMSException {
        return msg.getObjectProperty(VALUE_PROPERTY);
    }

    private byte[] serialize(final Object object) throws IOException {
        ByteArrayOutputStream bOutputStream = new ByteArrayOutputStream();
        ObjectOutputStream objectOutputStream = new ObjectOutputStream(bOutputStream);
        objectOutputStream.writeObject(object);
        objectOutputStream.flush();
        return bOutputStream.toByteArray();
    }

    private Object unserialize(final BytesMessage bytesMessage)
    throws IOException, JMSException, ClassNotFoundException {
        byte[] bytes = new byte[(int) bytesMessage.getBodyLength()];
        bytesMessage.readBytes(bytes);
        ObjectInputStream stream = new ObjectInputStream(
                new ByteArrayInputStream(bytes)) {
            @Override
            protected Class<?> resolveClass(final ObjectStreamClass desc) throws IOException, ClassNotFoundException {
                String name = desc.getName();
                return Class.forName(name, false, getClass().getClassLoader());
            }
        };
        try {
            return stream.readObject();
        } finally {
            stream.close();
        }
    }

    private String getExtTopicJNDIName() {
        return ((Config) getConfig()).getExtTopicJNDIName();
    }

    private String getLocalTopicJNDIName() {
        return ((Config) getConfig()).getLocalTopicJNDIName();
    }

    private String getLocalTopicName() {
        return ((Config) getConfig()).getLocalTopicName();
    }

    private String getExtProviderURL() {
        return ((Config) getConfig()).getExtProviderURL();
    }

    private boolean isClustered() {
        return ((Config) getConfig()).isClustered();
    }

    private boolean isHierarchical() {
        return ((Config) getConfig()).isHierarchical();
    }

    private boolean isClusterViewShared() {
        return ((Config) getConfig()).isClusterViewShared();
    }

    private String getObjectNameToListen() {
        return ((Config) getConfig()).getObjectNameToListen();
    }

    private String getJmsProviderName() {
        return ((Config) getConfig()).getJmsProviderName();
    }

    private String getConnectionFactoryName() {
        return ((Config) getConfig()).getConnectionFactoryName();
    }

    public long getPeriodToBeat() {
        return ((Config) getConfig()).getPeriodToBeat();
    }

    private long getPeriodToCleanUp() {
        return ((Config) getConfig()).getPeriodToCleanUp();
    }

    private long getWaitingBeatNumber() {
        return ((Config) getConfig()).getWaitingBeatNumber();
    }

    private boolean isPublisher() {
        return ((Config) getConfig()).isPublisher();
    }

    private boolean isSubscriber() {
        return ((Config) getConfig()).isSubscriber();
    }

    private String getDomainName() {
        return ((Config) getConfig()).getDomainName();
    }

    /**
     * @return
     */
    private short getDomainPort() {
        return ((Config) getConfig()).getDomainPort();
    }

    public ClusterView getClusterView() {
        return clusterView;
    }


    /**
     * Heart beat to notify the other servers that I stay alive.
     */
    private class Heartbeat implements Runnable {

        public void run() {
            Session session;
            MessageProducer producer;
            Message msg;
            try {
                session = topicConnection.createSession(false, Session.AUTO_ACKNOWLEDGE);
                producer = session.createProducer(topic);
                msg = session.createMessage();
                msg.setStringProperty(PROVIDER_URL_PROPERTY, getLocalServer().getProviderURL());
            } catch (JMSException e) {
                logger.error("Cannot init the producer for the heartbeat!", e);
                throw new RuntimeException(
                        "Cannot init the producer for the heartbeat!", e);
            }
            while(isImplementationAvailable()) {
                try {
                    producer.send(msg);
                } catch (JMSException e) {
                    logger.warn("Cannot send a heartbeat!", e);
                }
                try {
                    Thread.sleep(getPeriodToBeat());
                } catch (InterruptedException e) {
                    // Restore interrupted status
                    Thread.currentThread().interrupt();
                }
            }
        }

    }

    /**
     * Periodically clean the cluster view.
     */
    private class ClusterViewCleaner implements Runnable, MessageListener {

        /**
         * Timestamps of the last messages received from the other servers.
         */
        private Map<String, Long> timestamps;


        public void run() {
            timestamps = new ConcurrentHashMap<String, Long>();
            try {
                Session session = topicConnection.createSession(false, Session.AUTO_ACKNOWLEDGE);
                MessageConsumer consumer = session.createConsumer(topic);
                consumer.setMessageListener(this);
            } catch (JMSException e) {
                logger.error("Cannot init the consumer for the cluster view cleaner!", e);
                throw new RuntimeException(
                        "Cannot init the consumer for the cluster view cleaner!", e);
            }
            while(isImplementationAvailable()) {
                try {
                    Thread.sleep(getPeriodToCleanUp());
                } catch (InterruptedException e) {
                    // Restore interrupted status
                    Thread.currentThread().interrupt();
                }
                logger.debug("Cleaning the cluster view...");
                cleanClusterView();
            }
        }

        private synchronized void cleanClusterView() {
            Set<String> providerUrls =
                new HashSet<String>(clusterView.getServerViews().keySet());
            for (String providerURL : providerUrls) {
                // Get the last timestamp
                Long timestamp = timestamps.get(providerURL);
                if(timestamp != null &&
                        timestamp + getPeriodToBeat() * getWaitingBeatNumber() < System.currentTimeMillis()) {
                    if(providerURL.equals(getLocalServer().getProviderURL())) {
                        logger.warn("The local registry doesn't anymore send heartbeats!", providerURL);
                    } else {
                        logger.debug("Removing {0}", providerURL);
                        try {
                            removeServerView(providerURL);
                        } catch (ServerNotFoundException e) {
                            logger.warn("Error when removing " + providerURL, e);
                        }
                    }
                }
            }
        }


        public void onMessage(final Message msg) {
            if(msg instanceof TextMessage) {
                try {
                    timestamps.put(msg.getStringProperty(PROVIDER_URL_PROPERTY), msg.getJMSTimestamp());
                } catch (JMSException e) {
                    logger.warn("Cannot extract the content of {0}", msg, e);
                }
            }
        }

    }

}
